@@ -1,9 +1,10 @@
 import json
 #import jsonical
 import time
+import socket
 
 from twisted.protocols.basic import LineOnlyReceiver
-from twisted.internet import defer, reactor
+from twisted.internet import defer, reactor, error
 from twisted.python.failure import Failure
 
 #import services
@@ -41,7 +42,7 @@ class Protocol(LineOnlyReceiver):
         return self.request_id
 
     def _get_ip(self):
-        return self.transport.getPeer().host
+        return self.proxied_ip or self.transport.getPeer().host
 
     def get_session(self):
         return self.session
@@ -50,11 +51,18 @@ class Protocol(LineOnlyReceiver):
         try:
             self.transport.setTcpNoDelay(True)
             self.transport.setTcpKeepAlive(True)
+            self.transport.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 120) # Seconds before sending keepalive probes
+            self.transport.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 1) # Interval in seconds between keepalive probes
+            self.transport.socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 5) # Failed keepalive probles before declaring other end dead
         except:
             # Supported only by the socket transport,
             # but there's really no better place in code to trigger this.
             pass
 
+        # Read settings.TCP_PROXY_PROTOCOL documentation
+        self.expect_tcp_proxy_protocol_header = self.factory.__dict__.get('tcp_proxy_protocol_enable', False)
+        self.proxied_ip = None # IP obtained from TCP proxy protocol
+        
         self.request_id = 0    
         self.lookup_table = {}
         self.event_handler = self.factory.event_handler()
@@ -76,12 +84,14 @@ class Protocol(LineOnlyReceiver):
         
         
     def connectionLost(self, reason):
-        if not self.on_disconnect.called:
+        if self.on_disconnect != None and not self.on_disconnect.called:
             self.on_disconnect.callback(self)
-        
+            self.on_disconnect = None
+ 
         stats.PeerStats.client_disconnected(self._get_ip())
         connection_registry.ConnectionRegistry.remove_connection(self)
-       
+        self.transport = None # Fixes memory leak (cyclic reference)
+ 
     def writeJsonRequest(self, method, params, is_notification=False):
         request_id = None if is_notification else self._get_id() 
         serialized = json.dumps({'id': request_id, 'method': method, 'params': params})
@@ -171,12 +181,33 @@ class Protocol(LineOnlyReceiver):
                 request_counter.finish()
                 return self.lineLengthExceeded(line)
             else:
-                self.lineReceived(line, request_counter)
+                try:
+                    self.lineReceived(line, request_counter)
+                except:
+                    request_counter.finish()
+                    log.exception("Processing of message failed")
+                    return error.ConnectionLost('Processing of message failed')
+                    
         if len(self._buffer) > self.MAX_LENGTH:
             request_counter.finish()
             return self.lineLengthExceeded(self._buffer)        
         
     def lineReceived(self, line, request_counter):
+        if self.expect_tcp_proxy_protocol_header:
+            # This flag may be set only for TCP transport AND when TCP_PROXY_PROTOCOL
+            # is enabled in server config. Then we expect the first line of the stream
+            # may contain proxy metadata.
+
+            # We don't expect this header during this session anymore
+            self.expect_tcp_proxy_protocol_header = False
+            
+            if line.startswith('PROXY'):
+                self.proxied_ip = line.split()[2]
+
+                # Let's process next line
+                request_counter.decrease()
+                return
+            
         try:
             message = json.loads(line)
         except:
